/* Shannon Chappel's Golum Project
This is the code for the first Double-Rainbow module, which is to be installed
to run many of the torso modules.

Circuit outline:
9.6V base power
30mA orange LEDs.
4 groups of 3 LEDs each, 3 LEDs in series
Desired lighting effect: Autonomous flicker

Circuit design notes:
All wires are grouped and prewired with current limiting resistors selected for 9.6V operation (8 AA cells).
May bring system power to 11.1V Lipoly. Add additional resistors to enable 11~14V operation.

Software notes:
Using the great BLINK4 library from
http://playground.arduino.cc/Code/MultiBlink
https://arduinocode.codeplex.com/releases/view/115258

Possibly borrowing from Candle flicker code from
http://inkofpark.wordpress.com/2013/12/23/arduino-flickering-candle/

Larson scanner code modified for the weapon core from http://www.blinkenlight.net

We're using the millis() function to avoid delay() whereever we can to save cycles.

Dave Hrynkiw
Solarbotics Ltd, April 2014
Released under a Creative-Commons, Attribution, Share-alike License 4.0
*/

//#include <stdint.h>

//Used for candle flicker effect in the body crystals
const int flickerLEDs[] = {3,5,6,9};
const int8_t ledCount=4;
int8_t ledSet=0;

float fs = 60.;
float ts = 1./fs; //ts  = 0.01666
unsigned long tFlickerLimit = (unsigned long) (1000.0*ts); // = 16.66667. mS?
unsigned long tFlicker;

//Used for the glowing cracks
unsigned long tcrackGlowLimit=100; //milliseconds between updates (100=every 1/10th of a second)
unsigned long tCrackGrow;
const int crackGlowLevels[] = {30,40,50,60,70,80,90,120,150,180,200,254,254,254,254}; //brightness levels we'll toggle through up'n'down
#define crackGlowLevelTotal 14 //number of elements in the array
int8_t crackGlowIndex=0; //Initialize the brightness level variable
boolean crackGlowDir=1;  //We going up (1), or down (0) the brightness chain?
int crackGlowPin=10;	//Pin 10 to control the crack glow

//Used for the Chest tubes
unsigned long tChestGlowLimit=80; //milliseconds between updates (100=every 1/10th of a second)
unsigned long tChestGrow;
const int chestGlowLevels[] = {30,30,30,30,30,30,35,40,45,50,55,60,65,70,75,80,85,90,95,100,125,135,145,255}; //brightness levels we'll toggle through up'n'down
#define chestGlowLevelTotal 23 //number of elements in the array
int8_t chestGlowIndex=0; //Initialize the brightness level variable
boolean chestGlowDir=1;  //We going up (1), or down (0) the brightness chain?
int chestGlowPin=11;	//Pin 11	 to control the chest glow

//Used for checkBreaker
#define crystalBreakerPin 4
#define crackBreakerPin 7
#define chestBreakerPin 8
const int breaker[]={crystalBreakerPin, crackBreakerPin, chestBreakerPin};
boolean breakCrystals = 0;	//Disable Body crystal array
boolean breakCracks = 0;	//Disable Body Cracks
boolean breakChest=0;		//Disable Chest tubes


int16_t nextrand(){						//Used for candle flicker effect in the body crystals
	// Construct a nearly Normal random number by adding together four
	// uniformly distributed random values. The range was chosen to
	// provide the correct variance at the output of the filter.
	static const int numToAvg = 4;
	int16_t accum = 0;
	for( int i = 0;
	i < numToAvg;
	i++, accum += random( -616, 616));

	return accum / numToAvg;
}

unsigned long timeSince_ms( unsigned long tcomp){						//Used for candle flicker effect in the body crystals
	// This function deals with wrap-around on the millis() clock.
	// It can only handle a single wrap, of course
	unsigned long tnow = millis();
	
	if( tnow < tcomp)
	return( tnow + 4294967295 - tcomp);
	
	return( tnow - tcomp);
}

int8_t iirtwostage( int16_t imp){						//Used for candle flicker effect in the body crystals
	// The actual implementation of the two-stage filter. The
	// coefficients were designed in the Python code ecandle.py.
	static int16_t s[3];
	const static int16_t a[] = { 64, -56,  20};
	const static int16_t b[] = {  7,  14,   7};
	const static int16_t norm = 128; // Set normh to 2*a[0]
	const static int16_t offset = 172;

	int16_t y;

	s[0] = (imp * a[0]
	- a[0] * s[1]
	- a[2] * s[2])/a[0];
	y = b[0]*s[0]
	+ b[1]*s[1]
	+ b[2]*s[2];

	s[2] = s[1];
	s[1] = s[0];
	
	y /= norm;
	y += offset;
	y = y > 255 ? 255 : y;
	y = y < 0 ? 0 : y;
	return( int8_t(y));
}

void setup(){

	for (int thisFlickerLed = 0; thisFlickerLed < ledCount; thisFlickerLed++) { //Used for candle flicker effect in the body crystals
		pinMode(flickerLEDs[thisFlickerLed], OUTPUT);
	}
	pinMode(crackGlowPin, OUTPUT);
	tFlicker = millis();
	tCrackGrow = millis();
	tChestGrow = millis();

	//Setup the breaker pins
	pinMode(crystalBreakerPin, INPUT_PULLUP);
	pinMode(crackBreakerPin, INPUT_PULLUP);
	pinMode(chestBreakerPin, INPUT_PULLUP);

}

void loop(){

	if (digitalRead(crystalBreakerPin) == LOW) {breakCrystals = true;} else {breakCrystals = false;}
	if (digitalRead(crackBreakerPin) == LOW) {breakCracks = true;} else {breakCracks = false;}
	if (digitalRead(chestBreakerPin) == LOW) {breakChest = true;} else {breakChest = false;}

	bodyCrystals();
	crackGlow();
	chestTubeGlow();

}


void bodyCrystals(){

	if (breakCrystals==false) {	//If breaker is thrown to turn off crystals...
		for (int i=0; i < ledCount; i++){	//Go thru them all, and turn them off
			analogWrite(flickerLEDs[i], 0);
		}
	}
	else
	{

		//Used for candle flicker effect in the body crystals
		int16_t v;
		int8_t x;
		// Flicker the flickering LED
		if( timeSince_ms( tFlicker) > tFlickerLimit){
			//Cycle it for 4 sets of LEDS

			for (int thisFlickerLed = ledSet; thisFlickerLed < ledCount; thisFlickerLed=thisFlickerLed+2) {
				v = nextrand(  );
				v = iirtwostage( v );
				analogWrite(flickerLEDs[thisFlickerLed], v);

			}
			if (ledSet==0)
			{ledSet=1;}
			else
			{ledSet=0;}
			tFlicker = millis();

		}
	}
	// End of body crystal flicker effect
}

void crackGlow(){

	if (breakCracks==false) {	//If breaker is thrown to turn off crystals...
		analogWrite(crackGlowPin, 0);	
	}
	else
	{
		if( timeSince_ms(tCrackGrow) > tcrackGlowLimit){
			if( crackGlowDir == true) { //We getting brighter (1)?
				analogWrite(crackGlowPin,crackGlowLevels[crackGlowIndex]); //Light that sucker up according the value in the array
				crackGlowIndex++; //increment the array pointer for next time
			}
			else
			{ //No? Must be getting dimmer...
				analogWrite(crackGlowPin,crackGlowLevels[crackGlowIndex]); //Light that sucker up according the value in the array
				crackGlowIndex--; //decrement the array pointer for next time
			}

			if( crackGlowIndex==crackGlowLevelTotal) {crackGlowDir=0;}	//Are we at either limit of the array (then reverse direction)
			else if( crackGlowIndex==0){crackGlowDir=1;}
			tCrackGrow = millis();
		}
	}
}

void chestTubeGlow(){
	if (breakChest==false) {	//If breaker is thrown to turn off crystals...
		analogWrite(chestGlowPin, 0);
	}
	
	else
	{
		if( timeSince_ms(tChestGrow) > tChestGlowLimit){
			if( chestGlowDir == true) { //We getting brighter (1)?
				analogWrite(chestGlowPin,chestGlowLevels[chestGlowIndex]); //Light that sucker up according the value in the array
				chestGlowIndex++; //increment the array pointer for next time
			}
			else
			{ //No? Must be getting dimmer...
				analogWrite(chestGlowPin,chestGlowLevels[chestGlowIndex]); //Light that sucker up according the value in the array
				chestGlowIndex--; //decrement the array pointer for next time
			}

			if( chestGlowIndex==chestGlowLevelTotal) {chestGlowDir=0;}	//Are we at either limit of the array (then reverse direction)
			else if( chestGlowIndex==0){chestGlowDir=1;}
			tChestGrow = millis();
		}
	}
}
